class ZZT
inherit
	SDL_CONSTANT
	DIRECTION
	KNOWS_MASTER
	READ_TOOL
creation make, experimental
feature
	world : WORLD
	sdl : SDL
	chr : CHRTOOL

	player : PLAYER

	board : MAP

	make is
	local
		in_file : STD_FILE_READ
	do
		if argument_count < 1 then
			io.put_string("Please supply .zzt filename%N")
			die_with_code(exit_failure_code)
		else
			!!in_file.connect_to(argument(1))
			if not in_file.is_connected then
				std_error.put_string("Error opening ")
				std_error.put_string(argument(1))
				std_error.put_string("%N")
				die_with_code(exit_failure_code)
			end
		end

		dungeon_master.put_zzt(Current)
		!!chr.make
		!!world.make
		zzt_read_prepare(in_file)
		if world.load_world then
			io.put_string("Starting...%N")
			play_it
		else
			std_error.put_string("Error loading ZZT world.%N")
		end
	end

	play_it is
	do
		!!sdl.make

		board := world.board_list.item(world.start_board)
		!!player.make(board.start_square)
		player.init(world)

		board := world.board_list.item(0)
		demo_loop

		board := world.board_list.item(world.start_board)
		main_loop
	end

	--XXX: bad design: I'll change this later
	demo_loop is
	local
		i : INTEGER
		j : INTEGER
		list : LINKED_LIST[THING]
		j1, j2 : INTEGER
		finished : BOOLEAN
	do
		draw_map
		sdl.set_tick_interval(100)
		sdl.tick_reset
		from
		until finished
		loop
			from
			until sdl.tick_ready
			loop
			end

			list := board.tick_list
			from j := list.lower
			until j > list.upper
			loop
				if not list.item(j).ticking then
					list.remove(j)
				else
					j := j + 1
				end
			end
			j2 := list.upper
			from j := list.lower
			until j > j2
			loop
				board.tick_list.item(j).tick
				j := j + 1
			end

			if sdl.key_ready then
				finished := true
			end
			draw_map
		end
	end

	main_loop is
	local
		i : INTEGER
		j : INTEGER
		list : LINKED_LIST[THING]
		j1, j2 : INTEGER
		key : INTEGER
		finished : BOOLEAN
	do
		draw_map
		sdl.set_tick_interval(100)
		sdl.tick_reset
		from
		until finished
		loop
			from
			until sdl.tick_ready
			loop
			end

			if sdl.key_ready then
				key := sdl.last_key
				--io.put_string("Key pressed: ")
				--io.put_integer(key)
				--io.put_string("%N")
				inspect key
				when sdlk_escape then
					finished := true
				when sdlk_up then
					move(north)
				when sdlk_down then
					move(south)
				when sdlk_left then
					move(west)
				when sdlk_right then
					move(east)
				when sdlk_space then
					shoot
				else
				end
			end

			list := board.tick_list
			from j := list.lower
			until j > list.upper
			loop
				if not list.item(j).ticking then
					list.remove(j)
				else
					j := j + 1
				end
			end
			j2 := list.upper
			from j := list.lower
			until j > j2
			loop
				board.tick_list.item(j).tick
				j := j + 1
			end

			draw_map

			if player.health <= 0 then
				--GAME OVER
				finished := true
			end
		end
	end

	draw_map is
	local
		x, y : INTEGER
		sq : SQUARE
	do
		sdl.start_drawing
		from y := 1
		until y > 25
		loop
			from x := 1
			until x > 60
			loop
				sq := board.board.item(x,y)
				if sq /= Void then
					draw_thing(sq.dominant, x, y)
				end
				x := x + 1
			end
			y := y + 1
		end

		draw_status
		sdl.finish_drawing
	end

	draw_thing(t : THING; x, y : INTEGER) is
	local
		i , j : INTEGER
		ox, oy : INTEGER
		fg, bg : INTEGER
		a : ARRAY2[BOOLEAN]
	do
		a := chr.table.item(t.chr)
		ox := (x - 1) * chr.width
		oy := (y - 1) * chr.height
		fg := t.fg
		bg := t.bg
		from j := 1
		until j > chr.height
		loop
			from i := 1
			until i > chr.width
			loop
				if a.item(i, j) then
					sdl.draw_pixel(ox + i, oy + j, fg)
				else
					sdl.draw_pixel(ox + i, oy + j, bg)
				end
				i := i + 1
			end
			j := j + 1
		end
	end

	status_left_edge : INTEGER is 65
	draw_status is
	local
		s : STRING
	do
		current_fg := 15
		current_bg := 0
		!!s.copy(" Health:")
		s.append(player.health.to_string)
		print_at(status_left_edge, 8, s)
		!!s.copy("   Ammo:")
		s.append(player.ammo.to_string)
		print_at(status_left_edge, 9, s)
		!!s.copy("Torches:")
		s.append(player.torch.to_string)
		print_at(status_left_edge, 10, s)
		!!s.copy("   Gems:")
		s.append(player.gems.to_string)
		print_at(status_left_edge, 11, s)
		!!s.copy("  Score:")
		s.append(player.score.to_string)
		print_at(status_left_edge, 12, s)
		!!s.copy("   Keys:")
		print_at(status_left_edge, 13, s)
	end

	print_at(x, y : INTEGER; s : STRING) is
	do
		print_at_pixel((x - 1) * chr.width, (y - 1) * chr.height, s)
	end
	print_at_pixel(x, y : INTEGER; s : STRING) is
	local
		i : INTEGER
	do
		from i := 1
		until i > s.count
		loop
			draw_chr(x + (i - 1) * chr.width, y, s.item(i).code)
			i := i + 1
		end
	end

	current_fg : INTEGER
	current_bg : INTEGER
	draw_chr(x, y : INTEGER; c : INTEGER) is
	local
		i , j : INTEGER
		a : ARRAY2[BOOLEAN]
	do
		a := chr.table.item(c)
		from j := 1
		until j > chr.height
		loop
			from i := 1
			until i > chr.width
			loop
				if a.item(i, j) then
					sdl.draw_pixel(x + i, y + j, current_fg)
				else
					sdl.draw_pixel(x + i, y + j, current_bg)
				end
				i := i + 1
			end
			j := j + 1
		end
	end

	move(dir : INTEGER) is
	require
		dir >= min_dir and dir <= max_dir
	local
		loc : LOCATION
	do
		player.send_player_touch(dir)
	end

	goto_board(b : INTEGER) is
		-- for cheating/debugging
	do
		board.put_start_square(player.square)
		board := world.board_list.item(b)
		player.move_to(board.start_square)
	end

	passage_to(b, c : INTEGER) is
	local
		list : LINKED_LIST[PASSAGE]
		x, y : INTEGER
		sq : SQUARE
		found : BOOLEAN
		source : MAP
		t : THING
	do
		board.put_start_square(player.square)
		board := world.board_list.item(b)
		from x := 1
		until found or else x > board.width
		loop
			from y := 1
			until found or else y > board.height
			loop
				t := board.board.item(x, y).dominant
				if t.kind.is_equal("passage") and then t.colour = c then
					found := true
				else
					y := y + 1
				end
			end
			x := x + 1
		end
		if found then
			sq := t.square
		else
			sq := board.start_square
		end
		pause_at(sq)
	end
	
	pause_at(sq : SQUARE) is
		-- pause mode
	local
		key : INTEGER
		done : BOOLEAN
		dir : INTEGER
	do
		player.leave_square
		sq.move_all_to(backup_square)
		player.enter_square(sq)
		draw_map
		sdl.start_drawing
		print_at(status_left_edge, 6, "Pausing...")
		draw_thing(player, sq.location.x, sq.location.y)
		sdl.finish_drawing
		from
		until done
		loop
			key := sdl.get_key
			inspect key
			when sdlk_up then
				dir := north
			when sdlk_down then
				dir := south
			when sdlk_left then
				dir := west
			when sdlk_right then
				dir := east
			else
				dir := direction_idle
			end
			if dir >= min_dir and dir <= max_dir then
				move(dir)
				if player.square /= sq then --has the player moved?
					done := true
				end
			end
		end
		backup_square.move_all_to(sq)
		print_at(status_left_edge - 1, 3, "          ")
		sdl.tick_reset
		print_at(status_left_edge, 6, "          ")
	end

	backup_square : SQUARE is
	once
		!!Result.make(Void, Void)
	end

	shoot is --(dir : INTEGER) is
	local
		b : BULLET
	do
		if board.max_shot = 0 then
			single_line("Can't shoot in this place!")
		elseif board.bullet_count < board.max_shot then
			if player.ammo <= 0 then
				single_line("You don't have any ammo!")
			else
				if player.square.adjacent(player.direction).dominant.can_adjacent_shoot then
					player.add_ammo(-1)
					!!b.make_shoot(player.square, player.direction, 0)
				end
			end
		end
	end

	scroll_height : INTEGER is 15
	scroll_cursor_height : INTEGER is 8

	show_scroll(lines : ARRAY[STRING]) is
	local
		i, j : INTEGER
		cursor_line : INTEGER
		finished : BOOLEAN
		key : INTEGER
		s : STRING
		n : INTEGER
		target : STRING
	do
		if lines.count = 1 then
			single_line(lines.first)
		else
			cursor_line := 1
			from
			until finished
			loop
				sdl.start_drawing
				print_scroll_border
				target := Void
				from
					i := cursor_line - scroll_cursor_height + 1
					j := 1
				until j > scroll_height
				loop
					if i >= 1 and then i <= lines.count then
						s := lines.item(i)
						if s.first = '!' then
							-- link
							n := s.index_of(';')
							if cursor_line = i then
								target := s.substring(2, n - 1)
							end
							s := s.substring(n + 1, s.count)
							print_scroll_choice(j, s)
						elseif s.first = '$' then
							-- undocumented: heading
							print_scroll_heading(j, s.substring(2, s.count))
						else
							-- normal text
							print_scroll_line(j, s)
						end
					elseif i = 0 or else i = lines.count + 1 then
						print_scroll_heading(j, "-----------------")
					else
						print_scroll_line(j, "")
					end
					i := i + 1
					j := j + 1
				end
				sdl.finish_drawing
				key := sdl.get_key
				inspect key
				when sdlk_escape then
					finished := true
				when sdlk_up then
					if cursor_line > 1 then
						cursor_line := cursor_line - 1
					end
				when sdlk_down then
					if cursor_line < lines.count then
						cursor_line := cursor_line + 1
					end
				when sdlk_pageup then
					cursor_line := cursor_line - (scroll_height - 1)
					if cursor_line < 1 then
						cursor_line := 1
					end
				when sdlk_pagedown then
					cursor_line := cursor_line + scroll_height - 1
					if cursor_line > lines.count then
						cursor_line := lines.count
					end
				when sdlk_return then
					if target /= Void then
						dungeon_master.current_object.goto(target)
					end
					finished := true
				when sdlk_space then
					finished := true
				else
				end
			end
		end
	end

	line_max : INTEGER is 42

	print_scroll_border is
	local
		s : STRING
	do
		current_fg := 12
		current_bg := 1
		print_at(8, scroll_cursor_height + 6, "%/175/")
		print_at(8 + line_max + 1, scroll_cursor_height + 6, "%/174/")
		current_fg := 14
		!!s.copy(dungeon_master.current_object.id)
		s.precede_multiple(' ', (line_max - s.count) // 2)
		s.extend_to_count(' ', line_max)
		print_at(9, 5, s)
	end
	
	print_scroll_line(row : INTEGER; line : STRING) is
	local
		s : STRING
	do
		!!s.copy(line)
		s.extend_to_count(' ', line_max)
		current_fg := 14
		print_at(9, row + 6, s)
	end

	print_scroll_heading(row : INTEGER; line : STRING) is
	local
		s : STRING
	do
		!!s.copy(line)
		s.precede_multiple(' ', (line_max - line.count) // 2)
		s.extend_to_count(' ', line_max)
		current_fg := 15
		print_at(9, row + 6, s)
	end
	
	print_scroll_choice(row : INTEGER; line : STRING) is
	local
		s : STRING
	do
		!!s.copy(line)
		s.precede_multiple(' ', 5)
		s.extend_to_count(' ', line_max)
		current_fg := 15
		print_at(9, row + 6, s)
		current_fg := 13
		print_at(11, row + 6, "%/16/")
	end

	single_line_y : INTEGER is 27
	single_line(m : STRING) is
		--XXX: needs to die after a while
	local
		s : STRING
	do
		current_fg := 13
		current_bg := 0
		!!s.copy(m)
		s.precede_multiple(' ', (line_max - m.count) // 2)
		s.extend_to_count(' ', line_max)
		print_at(9, single_line_y, s)
	end

	-- EXPERIMENTAL: tiles and viewports
	experimental is
	local
		in_file : STD_FILE_READ
	do
		if argument_count < 1 then
			io.put_string("Please supply .zzt filename%N")
			die_with_code(exit_failure_code)
		else
			!!in_file.connect_to(argument(1))
			if not in_file.is_connected then
				std_error.put_string("Error opening ")
				std_error.put_string(argument(1))
				std_error.put_string("%N")
				die_with_code(exit_failure_code)
			end
		end

		dungeon_master.put_zzt(Current)
		!!chr.make
		!!world.make
		zzt_read_prepare(in_file)
		if world.load_world then
			experimental_play_it
		else
			std_error.put_string("Error loading ZZT world.%N")
		end
	end

	tile_width : INTEGER is 32
	tile_height : INTEGER is 32

	view_width : INTEGER is 10
	view_height : INTEGER is 10

	draw_viewport is
	local
		i, j : INTEGER
		x, y : INTEGER
		sq : SQUARE
		blank : THING
	do
		!FLOOR!blank.make(Void)
		sdl.start_drawing
		from j := 1
			y := player.square.location.y - view_height // 2
			if y < 1 then y := 1 end
		until j > view_height
		loop
			from i := 1
				x := player.square.location.x - view_width // 2
				if x < 1 then x := 1 end
			until i > view_width
			loop
				draw_tile(blank, i, j)
				if x <= 60 and y <= 25 then
					sq := board.board.item(x, y)
					if sq /= Void then
						draw_thing(sq.dominant, i, j)
					end
				end
				x := x + 1
				i := i + 1
			end
			y := y + 1
			j := j + 1
		end
		sdl.finish_drawing
	end

	draw_tile(t : THING; x, y : INTEGER) is
	local
		i , j : INTEGER
		ox, oy : INTEGER
		fg, bg : INTEGER
		a : ARRAY2[BOOLEAN]
	do
		a := chr.table.item(t.chr)
		ox := x * tile_width
		oy := y * tile_height
		fg := t.fg
		bg := t.bg
		if t.generating_type.is_equal("FLOOR") then
			sdl.draw_image(ox, oy, floor_image)
		elseif t.generating_type.is_equal("PLAYER") then
			sdl.draw_image(ox, oy, player_image)
		elseif t.generating_type.is_equal("SLIDEREW") then
			sdl.draw_image(ox, oy, sliderew_image)
		elseif t.generating_type.is_equal("SLIDERNS") then
			sdl.draw_image(ox, oy, sliderns_image)
		else
			from j := 1
			until j > chr.height
			loop
				from i := 1
				until i > chr.width
				loop
					if a.item(i, j) then
						sdl.draw_pixel(ox + i, oy + j, fg)
					else
						sdl.draw_pixel(ox + i, oy + j, bg)
					end
					i := i + 1
				end
				j := j + 1
			end
		end
	end

	experimental_play_it is
	local
		i : INTEGER
		j : INTEGER
		list : LINKED_LIST[THING]
		j1, j2 : INTEGER
		key : INTEGER
		finished : BOOLEAN
	do
		!!sdl.make

		board := world.board_list.item(world.start_board)
		!!player.make(board.start_square)
		player.init(world)

		board := world.board_list.item(0)
		draw_map
		key := sdl.get_key
		board := world.board_list.item(world.start_board)
		clear_screen

		draw_viewport
		sdl.set_tick_interval(100)
		sdl.tick_reset
		from
		until finished
		loop
			from
			until sdl.tick_ready
			loop
			end

			list := board.tick_list
			from j := list.lower
			until j > list.upper
			loop
				if not list.item(j).ticking then
					list.remove(j)
				else
					j := j + 1
				end
			end
			j2 := list.upper
			from j := list.lower
			until j > j2
			loop
				board.tick_list.item(j).tick
				j := j + 1
			end

			if sdl.key_ready then
				key := sdl.last_key
				--io.put_string("Key pressed: ")
				--io.put_integer(key)
				--io.put_string("%N")
				inspect key
				when sdlk_escape then
					finished := true
				when sdlk_up then
					move(north)
				when sdlk_down then
					move(south)
				when sdlk_left then
					move(west)
				when sdlk_right then
					move(east)
				when sdlk_space then
					shoot
				else
				end
			end
			draw_viewport
		end
	end
	
	clear_screen is
	local
		i, j : INTEGER
	do
		sdl.start_drawing
		from i := 1
		until i > 600
		loop
			from j := 1
			until j > 400
			loop
				sdl.draw_pixel(i, j, 0)
				j := j + 1
			end
			i := i + 1
		end
		sdl.finish_drawing
	end

	player_image : POINTER is
	once
		Result := sdl.read_png("data/man1.png")
	end

	floor_image : POINTER is
	once
		Result := sdl.read_png("data/black.png")
	end

	sliderew_image : POINTER is
	once
		Result := sdl.read_png("data/sliderew.png")
	end

	sliderns_image : POINTER is
	once
		Result := sdl.read_png("data/sliderns.png")
	end

feature {PLAYER}
	board_change_check(b : MAP) is
	do
		if board /= b then
			board.put_start_square(player.square)
			board := b
		end
	end

end
